#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "eval.h"
#include "types.h"
#include "memory.h"
#include "builtins.h"
#include "read.h"
#include "compat.h"

const char *stt_version = VERSION;

/* used to bind variables after evaluation */
typedef struct {
	stt_obj_t *def;
	stt_obj_t *var;
	stt_obj_t *value;
} varset;

/* user function executor 
 * this is capable of executing multiple functions in place
 * (tail recursion optimisation)
 *
 * ctx:  hosting stt_varctx_t
 * obj:  code of the call - the list interpreted as a function invocation.
 * func: CAR of the list evaluated as a function
 * args: the CDR (list of arguments)
 * tail: whether the call was made in a tail position
 *
 * inhibit_macro_expansion: if true, return values of macros aren't evaluated again.
 */
stt_obj_t * stt_funcexec(stt_varctx_t *ctx, stt_obj_t *obj, stt_obj_t *func, stt_obj_t *args, const int tail, int inhibit_macro_expansion)
{
	stt_varctx_t *fctx = NULL;	/* ctx for function code evaluation */
	stt_varctx_t *ectx = ctx;		/* ctx for function argument evaluation */

	stt_varctx_t *octx = ctx, *cctx, *ictx;
	varset bindings[FUNC_MAXPARMS];

	int nbind, k;

	stt_s_mark mark;

	stt_obj_t *parm, *vars;
	stt_obj_t *i, *j, *first, *cur;
	stt_obj_t *o;
	stt_obj_t *ofunc = func;
	stt_obj_t *oobj = obj;

	stt_obj_t *dispatch = NULL;
	stt_obj_t **argeval = NULL;

	stt_obj_t *rtrn = NULL; /* ultimate return value for methods */

	int ii, tailcall = 0;

	int trace = 0, closure, noafter = 0;

	mark = stt_gc_prot_mark(ctx->stt);

handlecall:

	/* repeats in case of a tail call - also for next methods,
	 * before/after method execution */

	closure = 0;

	switch(func->type) {

	case STT_ST_FUNC:
		/* just a function call. */
		trace = func->d.func.trace;
		closure = func->d.func.closure ? 1 : 0;
		break;

	case STT_ST_GENERIC:
		/* generic function */

		dispatch = stt_create_dispatch(ctx->stt);
		mark = stt_gc_prot_mark(ctx->stt);
		argeval = dispatch->d.clos_dispatch.argeval;

		dispatch->d.clos_dispatch.generic = func;

		/* evaluate all arguments */
		parm = args; ii = 0;
		while(stt_nextparm(&i, &parm)) {
			o = stt_eval(tailcall ? fctx : ectx, i);
			if(o->type == STT_ST_ERROR) {
				if(fctx && !fctx->ref) stt_varctx_free(fctx);
				stt_gc_prot_free(ctx->stt, mark, o);
				return o;
			}

			stt_gc_protlist_add(ctx->stt, o);

			argeval[ii++] = o;

			if(ii == FUNC_MAXPARMS) {
				if(fctx && !fctx->ref) stt_varctx_free(fctx);
				stt_gc_prot_free(ctx->stt, mark, NULL);
				return stt_create_error(ctx->stt, STT_SE_PNUM, "Maximum number of arguments for generic exceeded");
			}
		}

		argeval[ii] = NULL;

		/* try dispatch to a method */

		dispatch->d.clos_dispatch.method = stt_method_dispatch(ctx->stt, dispatch->d.clos_dispatch.generic, dispatch->d.clos_dispatch.argeval, 0);
		if(!dispatch->d.clos_dispatch.method) {
			if(fctx && !fctx->ref) stt_varctx_free(fctx);
			stt_gc_prot_free(ctx->stt, mark, NULL);
			return stt_create_error(ctx->stt, STT_SE_PARM, "Method dispatch failed");
		}

		if(dispatch->d.clos_dispatch.method->d.clos_method.place == CL_METHOD_PRIMARY) {
			/* try to run before methods before primary method. */

			if(dispatch->d.clos_dispatch.generic->d.clos_generic.m_before->type == STT_ST_CONS) {
				func = dispatch->d.clos_dispatch.generic->d.clos_generic.m_before->d.cons.car;
				goto handlecall;
			}
		}

		func = dispatch->d.clos_dispatch.method;
		goto handlecall;

		break;

	case STT_ST_METHOD:
		/* don't check for the function being a macro later on..
		 * it's not even a function */
		inhibit_macro_expansion = 1;

		switch(func->d.clos_method.place) {
		case CL_METHOD_BEFORE:
			/* executing a :before method. */

			if(!stt_is_applicable(ctx->stt, func, argeval))  {
				/* not applicable, look at next method */
				func = func->d.clos_method.next;

				if(func == NULL) {
					/* out of :before methods, jump to primary method. */
					func = dispatch->d.clos_dispatch.method;
				}

				goto handlecall;
			}

			break;

		case CL_METHOD_AFTER:
			/* executing an :after method. */

			if(!stt_is_applicable(ctx->stt, func, argeval)) {
				/* not applicable. look at next method */
				func = func->d.clos_method.next;

				if(func == NULL) {
					/* out of :after methods. return */

					return rtrn;
				}

				goto handlecall;
			}

			break;

		case CL_METHOD_AROUND:
			break;

		case CL_METHOD_PRIMARY:
			break;
		}

		break;
	
	case STT_ST_DISPATCH:

		/* call-next-method */

		dispatch = func;
		argeval = dispatch->d.clos_dispatch.argeval;

		/* we can set dispatch.method here (it's saved and restored
		 * in the call-next-method builtin...) */

		/* the old method gets passed in obj. sorry for that */

		func = dispatch->d.clos_dispatch.method = obj->d.clos_method.next;
		if(!func) {
			if(fctx && !fctx->ref) stt_varctx_free(fctx);
			stt_gc_prot_free(ctx->stt, mark, NULL);
			return stt_create_error(ctx->stt, STT_SE_WFUNC, "call-next-method called and there's no method available");
		}

		noafter = 1;

		if(obj->d.clos_method.place == CL_METHOD_AROUND && func->d.clos_method.place == CL_METHOD_PRIMARY) {
			/* get first applicable primary method */

			noafter = 0;

			func = dispatch->d.clos_dispatch.method = stt_method_dispatch(ctx->stt, dispatch->d.clos_dispatch.generic, argeval, 1);
			if(!func) {
				if(fctx && !fctx->ref) stt_varctx_free(fctx);
				stt_gc_prot_free(ctx->stt, mark, NULL);
				return stt_create_error(ctx->stt, STT_SE_PARM, "call-next-method called from :around with no primary method available");
			}

			/* try to run before methods before primary method. */

			if(dispatch->d.clos_dispatch.generic->d.clos_generic.m_before->type == STT_ST_CONS) {
				func = dispatch->d.clos_dispatch.generic->d.clos_generic.m_before->d.cons.car;
				goto handlecall;
			}
		}

		goto handlecall;

	default:
		abort(); /* doesn't happen. */
	}


	/* in case of tracing enabled, draw trace output */
	if(trace) {
		long ii;
		for(ii=0; ii<=ctx->stt->trace_depth; ii++) {
			o = stt_stream_stdout_write(ctx, ' ');
			if(o->type == STT_ST_ERROR) return o;
		}
		o = stt_stream_printf(ctx, "%c> ", tailcall ? '*' : '=');
		if(o->type == STT_ST_ERROR) return o;
		o = stt_printobj(ctx, obj ? obj->d.cons.car : NULL);
		if(o->type == STT_ST_ERROR) return o;
		o = stt_stream_printf(ctx, " (");
		if(o->type == STT_ST_ERROR) return o;
	}

	if(tailcall) {
		/* ectx is the context in which function arguments are evaluated,
		 * while in fctx the function code gets executed.
		 * in case of a tail call, the current fctx becomes ectx
		 */
		ectx = fctx;

		if(closure) {
			/* if it's not done already, make the current fctx into
			 * a first class object - this way the GC can reclaim it
			 * when necessary. */
			if(!fctx->ref) fctx->ref=stt_create_obj_varctx(ctx->stt, fctx);
		}
	} else {
		if(!closure) {
			fctx = stt_varctx_create(ctx->stt, ctx, FUNC_LOCALCTX_SIZE);
		}

		if(trace)
			ctx->stt->trace_depth ++;
	}

	if(closure) {
		fctx = stt_varctx_create(ctx->stt, NULL, FUNC_LOCALCTX_SIZE);
		cctx = NULL;

		/* link all varctxes described in the closure objects as parents
		 * to the new one created for this function invocation. */

		for(ictx = func->d.func.closure->d.varctx.ctx; ictx->parent; ictx = ictx->parent) {
			if(cctx) {
				cctx->parent = stt_varctx_create(ctx->stt, ctx, 0);
				cctx = cctx->parent;
			} else {
				cctx = stt_varctx_create(ctx->stt, ctx, 0);
				fctx->parent = cctx;
			}

			cctx->hash = ictx->hash;
			cctx->ns = ictx->ns;
			cctx->nsm = ictx->nsm;
			cctx->defer = 1;
			cctx->ref = stt_create_obj_varctx(ctx->stt, cctx);
		}
	}

	nbind = 0;

	if(func->type == STT_ST_METHOD) {
		/* bind method arguments */

		for(ii=0; ii<func->d.clos_method.argn; ii++) {
			if(!argeval[ii]) {
				if(!fctx->ref) stt_varctx_free(fctx);
				i = stt_create_error(ctx->stt, STT_SE_PARM, "Out of specified method arguments (bad call-next-method?)");
				stt_gc_prot_free(ctx->stt, mark, i);
				if(trace) ctx->stt->trace_depth --;
				return i;
			}

			stt_varctx_set_s(fctx, func->d.clos_method.args[ii].name->d.symbol.sdbm, func->d.clos_method.args[ii].name->d.symbol.id, argeval[ii]);
		}

		/* fill in bindings[] from method miscargs */
		vars = func->d.clos_method.miscargs;

		/* set _DISPATCH-INFO */
		stt_varctx_set_s(fctx, ctx->stt->dispatch_info_sdbm, ctx->stt->dispatch_info_id, dispatch);
	} else {
		/* bindings[] source is the standard function argument list */
		vars = func->d.func.parm;
	}

	/* map variables */
	parm = args;
	first = cur = NULL;

	/* fill in bindings array with function argument list */
	ii = 0;
	while(stt_nextparm(&i, &vars)) {
		if(ii == FUNC_MAXPARMS) {
			if(!fctx->ref) stt_varctx_free(fctx);
			i = stt_create_error(ctx->stt, STT_SE_PARM, "Function has too many arguments");
			stt_gc_prot_free(ctx->stt, mark, i);
			if(trace) ctx->stt->trace_depth --;
			/*if(closure) ctx->stt->gc_disabled--;*/
			return i;
		}

		if(i->type == STT_ST_CONS) {
			if(i->d.cons.cdr->type == STT_ST_CONS)
				bindings[ii].def = i->d.cons.cdr->d.cons.car;
			else
				bindings[ii].def = i->d.cons.cdr;
			bindings[ii].var = i->d.cons.car;
		} else {
			bindings[ii].def = NULL;
			bindings[ii].var = i;
		}

		bindings[ii].value = NULL;

		ii++;
	}
	nbind = ii;

	/* traverse call arguments */

	if(func->type == STT_ST_METHOD) {
		/* copy values from argeval to bindings */

		int l = 0;
		for(ii = func->d.clos_method.argn; argeval[ii]; ii++) {
			if(argeval[ii]->type == STT_ST_SYMBOL && argeval[ii]->d.symbol.colon) {
				/* key-value pair */

				i = argeval[ii];

				/* get value */
				ii ++;
				if(!argeval[ii]) {
					if(!fctx->ref) stt_varctx_free(fctx);
					i = stt_create_error(ctx->stt, STT_SE_PARM, "Can't have a key/value pair without a value");
					stt_gc_prot_free(ctx->stt, mark, i);
					if(trace) ctx->stt->trace_depth --;
					return i;
				}

				j = argeval[ii];

				/* find variable slot for key */
				for(k=0; k<nbind; k++)
					if(bindings[k].var->d.symbol.id == i->d.symbol.id)
						break;

				if(k == nbind) {
					/* invalid key */
					if(!fctx->ref) stt_varctx_free(fctx);
					i = stt_create_error(ctx->stt, STT_SE_PARM, "Invalid parameter key");
					stt_gc_prot_free(ctx->stt, mark, i);
					if(trace) ctx->stt->trace_depth --;
					return i;
				}

				bindings[k].value = j;
			} else {
				/* by-position argument */

				i = argeval[ii];

				if(l == nbind) {
					/* argument list exhausted, append this to the _rest list */
					if(!cur) {
						first = cur = stt_create_cons(ctx->stt, i, stt_create_nil(ctx->stt));
					} else {
						cur->d.cons.cdr = stt_create_cons(ctx->stt, i, stt_create_nil(ctx->stt));
						cur = cur->d.cons.cdr;
					}
				} else {
					bindings[l].value = i;
					l++;
				}
			}
		}

	} else {

		/* evaluate function argument list, as usual */

		ii = 0;
		while(stt_nextparm(&i, &parm)) {
			if(i->type == STT_ST_SYMBOL && i->d.symbol.colon && !func->d.func.macro) {
				/* key-value pair */

				/* already got key, get value */
				if(!stt_nextparm(&j, &parm)) {
					if(!fctx->ref) stt_varctx_free(fctx);
					i = stt_create_error(ctx->stt, STT_SE_PARM, "Can't have a key/value pair without a value");
					stt_gc_prot_free(ctx->stt, mark, i);
					if(trace) ctx->stt->trace_depth --;
					/*if(closure) ctx->stt->gc_disabled--;*/
					return i;
				}

				/* evaluate argument */
                j = stt_eval(ectx, j);
                if(j->type == STT_ST_ERROR) {
                    if(!fctx->ref) stt_varctx_free(fctx);
                    stt_gc_prot_free(ctx->stt, mark, j);
                    if(trace) ctx->stt->trace_depth --;
					/*if(closure) ctx->stt->gc_disabled--;*/
                    return j;
                }
                stt_gc_protlist_add(ctx->stt, j);

				/* find variable slot for key */
				for(k=0; k<nbind; k++)
					if(bindings[k].var->d.symbol.id == i->d.symbol.id)
						break;

				if(k == nbind) {
					/* invalid key */

					if(!fctx->ref) stt_varctx_free(fctx);
					i = stt_create_error(ctx->stt, STT_SE_PARM, "Invalid parameter key");
					stt_gc_prot_free(ctx->stt, mark, i);
					if(trace) ctx->stt->trace_depth --;
					/*if(closure) ctx->stt->gc_disabled--;*/
					return i;
				}

				if(trace) {
					/* if tracing is enabled, draw trace output. */
					o = stt_stream_stdout_write(ctx, ' ');
					if(o->type == STT_ST_ERROR) return o;
					o = stt_printobj(ctx, i);
					if(o->type == STT_ST_ERROR) return o;
					o = stt_stream_stdout_write(ctx, ' ');
					if(o->type == STT_ST_ERROR) return o;
					o = stt_printobj(ctx, j);
				}

				bindings[k].value = j;

			} else {
				/* by-position argument */

				/* evaluate argument */
				if(!func->d.func.macro) { /* macro functions don't get their parameters evaluated */
					i = stt_eval(ectx, i);
					if(i->type == STT_ST_ERROR) {
						if(!fctx->ref) stt_varctx_free(fctx);
						stt_gc_prot_free(ctx->stt, mark, i);
						if(trace) ctx->stt->trace_depth --;
						/*if(closure) ctx->stt->gc_disabled--;*/
						return i;
					}
					stt_gc_protlist_add(ctx->stt, i);
				}

				if(trace) {
					/* if tracing is enabled, draw trace output. */
					o = stt_stream_stdout_write(ctx, ' ');
					if(o->type == STT_ST_ERROR) return o;
					o = stt_printobj(ctx, i);
					if(o->type == STT_ST_ERROR) return o;
				}

				if(ii == nbind) {
					/* argument list exhausted, append this to the _rest list */
					if(!cur) {
						first = cur = stt_create_cons(ctx->stt, i, stt_create_nil(ctx->stt));
					} else {
						cur->d.cons.cdr = stt_create_cons(ctx->stt, i, stt_create_nil(ctx->stt));
						cur = cur->d.cons.cdr;
					}
				} else {
					bindings[ii].value = i;
					ii++;
				}
			}
		}
	}

	if(tailcall && fctx->ref) {
		/* fctx is referenced by a closure, we can't reuse it */
		fctx = stt_varctx_create(ctx->stt, ctx, FUNC_LOCALCTX_SIZE);
	}

	/* set existent bindings in the stt_varctx_t */
	for(k=0; k<nbind; k++)
		if(bindings[k].value)
			stt_varctx_set_s(fctx, bindings[k].var->d.symbol.sdbm, bindings[k].var->d.symbol.id, bindings[k].value);

	/* check for unset slots, evaluate default values */
	for(k=0; k<nbind; k++) {
		if(!bindings[k].value) {
			if(!bindings[k].def) {
				/* no default value for parameter */
				if(!fctx->ref) stt_varctx_free(fctx);
				i = stt_create_error_with_obj(ctx->stt, STT_SE_PNUM, "Insufficient arguments when calling user function", func);
				stt_gc_prot_free(ctx->stt, mark, i);
				if(trace) ctx->stt->trace_depth --;
				/*if(closure) ctx->stt->gc_disabled--;*/
				return i;
			} else {
				/* evaluate and bind default parameter */

				i = stt_eval(fctx, bindings[k].def);
				if(i->type == STT_ST_ERROR) {
					if(!fctx->ref) stt_varctx_free(fctx);
					stt_gc_prot_free(ctx->stt, mark, i);
					if(trace) ctx->stt->trace_depth --;
					/*if(closure) ctx->stt->gc_disabled--;*/
					return i;
				}

				stt_varctx_set_s(fctx, bindings[k].var->d.symbol.sdbm, bindings[k].var->d.symbol.id, i);
			}
		}
	}

	if(trace) {
		o = stt_stream_printf(ctx, " ) at %s:%d\n", obj ? obj->file : NULL, obj ? obj->line : 0);
		if(o->type == STT_ST_ERROR) return o;
	}

	/* bind _rest in the stt_varctx_t */
	stt_varctx_set_s(fctx, ctx->stt->rest_sdbm, ctx->stt->rest_id, first ? first : stt_create_nil(ctx->stt));

	/* bind <^ */
	stt_varctx_set_s(fctx, ctx->stt->recurse_sdbm, ctx->stt->recurse_id, func);

	/* execute function body */

	if(func->type == STT_ST_METHOD) {
		/* executing a method */


		parm = func->d.clos_method.code;
		j = NULL;

		while(stt_nextparm(&i, &parm)) {
			j = stt_eval(fctx, i);

			if(j->type == STT_ST_ERROR) {
				/* append location to trace list */
				if(!j->d.error.trace) {
					j->d.error.trace = stt_create_cons(ctx->stt, o = stt_create_error(ctx->stt, j->d.error.code, NULL), stt_create_nil(ctx->stt));
				} else {
					for(o = j->d.error.trace; o->d.cons.cdr->type == STT_ST_CONS; o = o->d.cons.cdr);
					o->d.cons.cdr = stt_create_cons(ctx->stt, stt_create_error(ctx->stt, j->d.error.code, NULL), stt_create_nil(ctx->stt));
					o = o->d.cons.cdr->d.cons.car;
				}
				o->d.error.obj = obj;
				o->d.error.file = obj->file;
				o->d.error.line = obj->line;

				if(!fctx->ref) stt_varctx_free(fctx);
				if(trace) ctx->stt->trace_depth --;
				stt_gc_prot_free(ctx->stt, mark, j);
				return j;
			}
		}
		if(!j) j = stt_create_nil(ctx->stt); /* empty functions exist.. */

	} else {

		/* executing an ordinary function */

		/*if(closure) stt_gc_protect_root(ctx);*/

		parm = func->d.func.exp;
		j = NULL;
		while(stt_nextparm(&i, &parm)) {
			j = stt_eval_(fctx, i, parm->type != STT_ST_CONS);

			if(j->type == STT_ST_DEFER) break; /* we got a tailcall :) */

			if(j->type == STT_ST_ERROR) {
				/* append location to trace list */
				if(!j->d.error.trace) {
					j->d.error.trace = stt_create_cons(ctx->stt, o = stt_create_error(ctx->stt, j->d.error.code, NULL), stt_create_nil(ctx->stt));
				} else {
					for(o = j->d.error.trace; o->d.cons.cdr->type == STT_ST_CONS; o = o->d.cons.cdr);
					o->d.cons.cdr = stt_create_cons(ctx->stt, stt_create_error(ctx->stt, j->d.error.code, NULL), stt_create_nil(ctx->stt));
					o = o->d.cons.cdr->d.cons.car;
				}
				o->d.error.obj = obj;
				o->d.error.file = obj->file;
				o->d.error.line = obj->line;

				if(!fctx->ref) stt_varctx_free(fctx);
				if(trace) ctx->stt->trace_depth --;
				/*if(closure) stt_gc_unprotect_root(ctx);*/
				stt_gc_prot_free(ctx->stt, mark, j);
				/*if(closure) ctx->stt->gc_disabled--;*/
				return j;
			}
		}
		if(!j) j = stt_create_nil(ctx->stt); /* empty functions exist.. */

		/*if(closure) stt_gc_unprotect_root(ctx);*/
		/*if(closure) ctx->stt->gc_disabled--;*/

		/* handle a tailcall */
		if(j->type == STT_ST_DEFER) {
			func = j->d.defer.func;
			args = j->d.defer.args;
			obj = j->d.defer.obj;
			tailcall = 1;

			stt_gc_prot_free(ctx->stt, mark, NULL);

			goto handlecall;
		}
	}

	/* finish function execution. */

	if(trace) {
		long ii;
		/* if tracing is enabled, draw trace output */
		for(ii = 0; ii < ctx->stt->trace_depth; ii++) {
			o = stt_stream_stdout_write(ctx, ' ');
			if(o->type == STT_ST_ERROR) return o;
		}
		o = stt_stream_printf(ctx, "<= ");
		if(o->type == STT_ST_ERROR) return o;
		o = stt_printobj(ctx, j);
		if(o->type == STT_ST_ERROR) return o;
		o = stt_stream_stdout_write(ctx, '\n');
		if(o->type == STT_ST_ERROR) return o;
		ctx->stt->trace_depth --;
	}

	if(!inhibit_macro_expansion) {
		if(ofunc->d.func.macro == 2) {
			/* this is an r-macro, we get to evaluate the results AGAIN! yay! */
			if(fctx->ref) stt_gc_protlist_add(ctx->stt, fctx->ref);
			j = stt_eval_(octx, j, tail);
		} else if(ofunc->d.func.macro & 1) {
			/* this is a macro. we replace the original s-expression with the
			 * results. */
			if(fctx->ref) stt_gc_protlist_add(ctx->stt, fctx->ref);
#ifdef REPORT_EXPANSION
			stt_printobj(octx, ofunc);
			printf(" <- macro %s:%d\n", oobj->file, oobj->line);
#endif

			if(ctx->stt->macro_annot) {
				stt_obj_t *n = stt_alloc_obj(ctx->stt);
				stt_deepcopy(ctx->stt, n, oobj);
				/* put (annot- ..) around macro results */
				j = stt_create_annot(ctx->stt, j, n);
			}

/*			char *f = oobj->file;
			int l = oobj->line;*/
			stt_freedata(oobj);
			stt_deepcopy(ctx->stt, oobj, j);
/*			oobj->file = f;
			oobj->line = l;*/
			j = stt_eval_(octx, oobj, tail);
		}
	}

	if(!fctx->ref) stt_varctx_free(fctx);

	stt_gc_prot_free(ctx->stt, mark, j);
	if(rtrn) stt_gc_protlist_add(ctx->stt, rtrn);

	if(func->type == STT_ST_METHOD) {
		/* progress with method execution */

		switch(func->d.clos_method.place) {
		case CL_METHOD_BEFORE:
			func = func->d.clos_method.next;
			if(!func) func = dispatch->d.clos_dispatch.method;
			goto handlecall;

		case CL_METHOD_PRIMARY:
			rtrn = j;

			if(!noafter && dispatch->d.clos_dispatch.generic->d.clos_generic.m_after->type == STT_ST_CONS) {
				func = dispatch->d.clos_dispatch.generic->d.clos_generic.m_after->d.cons.car;
				goto handlecall;
			}
			break;

		case CL_METHOD_AFTER:
			if((func = func->d.clos_method.next)) {
				goto handlecall;
			}

			j = rtrn;

			break;

		default: ;
		}
	}

	return j;
}

/* evaluate expression */
stt_obj_t *stt_eval_(stt_varctx_t *ctx, stt_obj_t *obj, const int tail)
{
	stt_var_t *var;
	stt_obj_t *func;
	stt_obj_t *i, *j;
	stt_s_mark mark;

	switch(obj->type) {
	case STT_ST_CONS:
		/* evaluate s-expression */

		if(ctx->stt->gc_allocd > ctx->stt->gc_alloc_limit)
			stt_gc_sweep(ctx->stt, ctx);

		func = stt_eval(ctx, obj->d.cons.car);
		if(func->type == STT_ST_ERROR) return func;


		switch(func->type) {
			case STT_ST_BUILTIN:
				/* builtin function call */

				ctx->stt->lastcall_file = obj->file;
				ctx->stt->lastcall_line = obj->line;

				/* optimise by replacing symbol in s-exp directly with builtin */
				if(obj->d.cons.car->type == STT_ST_SYMBOL &&
					obj->d.cons.car->d.symbol.id == func->d.builtin.symbolid &&
					!func->d.builtin.nosubst) {

					obj->d.cons.car = func;
				}

				mark = stt_gc_prot_mark(ctx->stt);

				i = func->d.builtin.handler(ctx, obj->d.cons.cdr, func->d.builtin.data, tail);
				if(i->type == STT_ST_ERROR) {
					if(!i->d.error.trace) {
						i->d.error.trace = stt_create_cons(ctx->stt, j = stt_create_error(ctx->stt, i->d.error.code, NULL), stt_create_nil(ctx->stt));
					} else {
						for(j = i->d.error.trace; j->d.cons.cdr->type == STT_ST_CONS; j = j->d.cons.cdr);
						j->d.cons.cdr = stt_create_cons(ctx->stt, stt_create_error(ctx->stt, i->d.error.code, NULL), stt_create_nil(ctx->stt));
						j = j->d.cons.cdr->d.cons.car;
					}
					j->d.error.obj = obj;
					j->d.error.file = obj->file;
					j->d.error.line = obj->line;
				}

				stt_gc_prot_free(ctx->stt, mark, i);
				return i;

			case STT_ST_FUNC:
				/* user function call */

				ctx->stt->lastcall_file = func->file;
				ctx->stt->lastcall_line = func->line;

				if(tail && !func->d.func.macro) {
					/* this is a tailcall!
					 * defer execution to parent funcexec */
					return stt_create_defer(ctx->stt, func, obj->d.cons.cdr, obj);
				}

				ctx->stt->recursion_limit --;
				if(!ctx->stt->recursion_limit) {
					ctx->stt->recursion_limit = 0;
					return stt_create_error_with_obj(ctx->stt, STT_SE_LIMIT, "Recursion limit exceeded", obj);
				}

				j = stt_funcexec(ctx, obj, func, obj->d.cons.cdr, tail, 0);

				ctx->stt->recursion_limit ++;
				return j;

			case STT_ST_GENERIC:
				/* generic function invocation */

				ctx->stt->lastcall_file = func->file;
				ctx->stt->lastcall_line = func->line;

				if(tail) {
					/* tail call, defer execution to parent funcexec */
					return stt_create_defer(ctx->stt, func, obj->d.cons.cdr, obj);
				}

				ctx->stt->recursion_limit --;
				if(!ctx->stt->recursion_limit) {
					ctx->stt->recursion_limit = 0;
					return stt_create_error_with_obj(ctx->stt, STT_SE_LIMIT, "Recursion limit exceeded", obj);
				}

				j = stt_funcexec(ctx, obj, func, obj->d.cons.cdr, tail, 0);

				ctx->stt->recursion_limit ++;
				return j;

			default:
				return stt_create_error_with_obj(ctx->stt, STT_SE_WFUNC, "Non-function in S-expression", obj);
		}
		break;

	case STT_ST_SYMBOL:
		if(obj->d.symbol.colon)
			return obj; /* keyword symbol */

		/* look up symbol */
		var = stt_varctx_get_s(ctx, obj->d.symbol.sdbm, obj->d.symbol.id);
		if(!var) {
			char buf[256];

			ctx->stt->lastcall_file = obj->file;
			ctx->stt->lastcall_line = obj->line;

			sprintf(buf, "Variable '%s' has no value", obj->d.symbol.name);
			obj = stt_create_error(ctx->stt, STT_SE_UNDEF, buf);
			return obj;
		}

		if(var->value->type == STT_ST_FUNC && (var->value->d.func.macro & 4)) {
			ctx->stt->lastcall_file = obj->file;
			ctx->stt->lastcall_line = obj->line;

			j = stt_funcexec(ctx, stt_create_cons(ctx->stt, var->value, stt_create_nil(ctx->stt)), var->value, stt_create_nil(ctx->stt), 0, 1);
			if(j->type == STT_ST_ERROR) return j;
			/*if(var->value->d.func.macro == 4) {
				stt_freedata(obj);
				stt_deepcopy(ctx->stt, obj, j);
				j = obj;
			}*/

			j = stt_eval_(ctx, j, 0);
			return j;
		}

		return var->value;

	case STT_ST_ANNOT:
		i = stt_eval_(ctx, obj->d.annot.val, tail);

		if(i->type == STT_ST_ERROR) {
			if(!i->d.error.trace) {
				i->d.error.trace = stt_create_cons(ctx->stt, j = stt_create_error(ctx->stt, i->d.error.code, NULL), stt_create_nil(ctx->stt));
			} else {
				for(j = i->d.error.trace; j->d.cons.cdr->type == STT_ST_CONS; j = j->d.cons.cdr);
				j->d.cons.cdr = stt_create_cons(ctx->stt, stt_create_error(ctx->stt, i->d.error.code, NULL), stt_create_nil(ctx->stt));
				j = j->d.cons.cdr->d.cons.car;
			}
			j->d.error.obj = obj->d.annot.annot;
			j->d.error.file = obj->d.annot.annot->file;
			j->d.error.line = obj->d.annot.annot->line;
		}

		return i;

	default:
		/* directly return atom */
		return obj;
	}

	return stt_create_error(ctx->stt, STT_SE_WTF, "Cannot evaluate object");
}

/* load and evaluate from fp */
stt_obj_t *stt_evalfp(stt_varctx_t *ctx, FILE *fp)
{
	int c, ol;
	stt_s_mark mark;
	stt_obj_t *obj = NULL;
	stt_obj_t *lobj = NULL;
	stt_read_state_t *rs;

	mark = stt_gc_prot_mark(ctx->stt);

	ol = ctx->stt->parsed_line;

	rs = stt_create_read_state(ctx->stt);

	for(;;) {
		c = fgetc(fp);
		obj = stt_read_ch(rs, c);
		if(obj && obj->type == STT_ST_ERROR) {
			stt_free_read_state(rs);
			stt_gc_prot_free(ctx->stt, mark, obj);
			return obj;
		}

		if(obj) {
			lobj = obj = stt_eval(ctx, obj);
			if(obj->type == STT_ST_ERROR) {
				if(obj->d.error.code == STT_SE_STOPREAD)
					obj = stt_create_nil(ctx->stt);
				stt_free_read_state(rs);
				stt_gc_prot_free(ctx->stt, mark, obj);
				return obj;
			}
		}

		if(c == EOF) break;
	}

	stt_free_read_state(rs);
	
	ctx->stt->parsed_line = ol;

	if(!lobj) lobj = stt_create_nil(ctx->stt);
	stt_gc_prot_free(ctx->stt, mark, lobj);
	return lobj;
}

/* load and evaluate from string */
stt_obj_t *stt_evalstr(stt_varctx_t *ctx, const char *script)
{
	int ol;
	stt_s_mark mark;
	stt_obj_t *obj = NULL;
	stt_obj_t *lobj = NULL;
	stt_read_state_t *rs;

	mark = stt_gc_prot_mark(ctx->stt);

	ol = ctx->stt->parsed_line;

	rs = stt_create_read_state(ctx->stt);

	for(;;script++) {
		int c = *script;
		if (c == 0) c = EOF;
		obj = stt_read_ch(rs, c);
		if(obj && obj->type == STT_ST_ERROR) {
			stt_free_read_state(rs);
			stt_gc_prot_free(ctx->stt, mark, obj);
			return obj;
		}

		if(obj) {
			lobj = obj = stt_eval(ctx, obj);
			if(obj->type == STT_ST_ERROR) {
				if(obj->d.error.code == STT_SE_STOPREAD)
					obj = stt_create_nil(ctx->stt);
				stt_free_read_state(rs);
				stt_gc_prot_free(ctx->stt, mark, obj);
				return obj;
			}
		}

		if(c == EOF) break;
	}

	stt_free_read_state(rs);
	
	ctx->stt->parsed_line = ol;

	if(!lobj) lobj = stt_create_nil(ctx->stt);
	stt_gc_prot_free(ctx->stt, mark, lobj);
	return lobj;
}


static char loadbuf[512];

/* load and evaluate file */
stt_obj_t *stt_evalfile(stt_varctx_t *ctx, const char *name)
{
	FILE *fp;
	stt_obj_t *r;
	char *of, *e;
	char buf[512];
	int l;

	if(name[0] == STT_PATHSEP[0] ||
		(strlen(name) > 3 && name[1] == ':' && name[2] == '\\')) {
		/* using absolute path */
		strcpy(buf, name);
		e = strrchr(buf, STT_PATHSEP[0]);
		*e = 0;
	} else {
		/* relative path */
		strcpy(buf, ctx->stt->load_path);
		for(e = buf, l = 0; *e; e++, l++);
		if(l >= 500)
			return stt_create_error(ctx->stt, STT_SE_PARM, "Load path too long");
		*e++ = STT_PATHSEP[0];
		l ++;
		buf[511] = 0;
		strncpy(e, name, 511-l);
		strcpy(loadbuf, buf);
		name = loadbuf;
		e = strrchr(buf, STT_PATHSEP[0]);
		if(e)
			*e = 0;
		else
			strcpy(buf, ".");
	}

	fp = fopen(name, "r");
	if(!fp)
		return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "evalfile: can't open file", STT_STR(ctx->stt, name));

	{
		char *old_load_path = ctx->stt->load_path;
		ctx->stt->load_path = buf;

		of = ctx->stt->file;
		stt_mark_miscfree(ctx->stt, ctx->stt->file = stt_strdup(name));
		r = stt_evalfp(ctx, fp);
		fclose(fp);
		ctx->stt->file = of;

		ctx->stt->load_path = old_load_path;
	}
	return r;
}

/* call (require 'system) */
stt_obj_t * stt_load_system(stt_varctx_t *ctx)
{
	return stt_eval(ctx, stt_create_cons(ctx->stt, stt_create_symbol(ctx->stt, "REQUIRE"), stt_create_cons(ctx->stt, stt_quoteobj(ctx->stt, stt_create_symbol(ctx->stt, "SYSTEM")), stt_create_nil(ctx->stt))));
}

/* register new error - returns new code */
int stt_regerrorcode(stt_ctx_t *stt, char *name)
{
	stt->errorcount ++;
	stt->errornames = (char *) stt_realloc(stt->errornames, stt->errorcount * 16);
	strcpy(stt->errornames + (stt->errorcount-1)*16, name);
	return stt->errorcount - 1;
}

/* returns name of error by id */
char *stt_errorname(stt_ctx_t *stt, int id)
{
	if(id >= stt->errorcount) return "UNKNOWN";
	return stt->errornames + id*16;
}
