#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "builtins.h"
#include "eval.h"
#include "types.h"
#include "read.h"
#include "compat.h"

/* create stt_read_state_t */
stt_read_state_t *stt_create_read_state(stt_ctx_t *stt)
{
	stt_read_state_t *s = stt_calloc(1, sizeof(stt_read_state_t));
	s->stt = stt;
	stt_read_reset(s);
	s->line = 1;
	s->col = 0;
	return s;
}

/* free current token data */
static void resettoken(stt_read_state_t *s)
{
	if(s->token) {
		free(s->token);
		s->token = NULL;
		s->buflen = s->strlen = 0;
	}
}

/* destroy stt_read_state_t */
void stt_free_read_state(stt_read_state_t *s)
{
	resettoken(s);
	free(s);
}

/* reset parser */
void stt_read_reset(stt_read_state_t *s)
{
	resettoken(s);
	s->state = STT_RS_NORM;
	s->stack = stt_create_nil(s->stt);
	s->out = stt_create_nil(s->stt);
	s->act = stt_create_nil(s->stt);
	s->ahead = 0;
	s->label = stt_create_nil(s->stt);
	s->stacklabel = -1;
}

/* append character to current token */
static void token_appch(stt_read_state_t *s, const char ch)
{
	if(!s->token) {
		s->token = (char *) stt_malloc(8);
		s->buflen = 8;
		s->strlen = 2;
		s->token[0] = ch;
		s->token[1] = 0;
		return;
	} else {
		if(s->strlen == s->buflen)
			s->token = (char *) stt_realloc(s->token, s->buflen *= 2);

		s->token[s->strlen] = 0;
		s->token[(s->strlen++)-1] = ch;
	}
}

/* append object to current list */
static int read_append(stt_read_state_t *s, stt_obj_t *obj)
{
	stt_obj_t *prv;
	int code;

	if(s->act->type == STT_ST_NIL) {
		/* list is currently empty */

		s->out = s->act = stt_create_cons(s->stt, obj, stt_create_nil(s->stt));
		if(s->stacklabel > -1) {
			s->label = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_int(s->stt, s->stacklabel), s->out), s->label);
			s->stacklabel = -1;
		}

		return 0;
	} else if(s->act->type == STT_ST_CONS && s->act->d.cons.cdr->type == STT_ST_NIL) {

		s->act->d.cons.cdr = stt_create_cons(s->stt, obj, stt_create_nil(s->stt));
		if(obj->type == STT_ST_ERROR) { /* appending list special */
			s->act = s->act->d.cons.cdr;
			return 0;
		}

		/* handling list specials */
		while(s->act->d.cons.car->type == STT_ST_ERROR) {
			prv = s->act->d.cons.car->d.error.obj;
			code = s->act->d.cons.car->d.error.code;

			/* list special: . */
			if(code == STT_SE_WTF) {
				if(prv->type != STT_ST_CONS) return 1;
				if(s->act->d.cons.cdr->type != STT_ST_CONS) return 2;
				prv->d.cons.cdr = s->act->d.cons.cdr->d.cons.car;
				s->act = prv;
				continue;
			}

			/* list special: ' */
			if(code == STT_SE_UNDEF) {
				if(s->act->d.cons.cdr->type != STT_ST_CONS) return 3;
				if(prv->type == STT_ST_NIL) {
					s->act = s->out = stt_create_cons(s->stt, stt_quoteobj(s->stt, s->act->d.cons.cdr->d.cons.car), stt_create_nil(s->stt));
					return 0;
				} else if(prv->type == STT_ST_CONS) {
					prv->d.cons.cdr = stt_create_cons(s->stt, stt_quoteobj(s->stt, s->act->d.cons.cdr->d.cons.car), stt_create_nil(s->stt));
					s->act = prv;
					continue;
				} else return 4;
			}

			/* list special: ` */
			if(code == STT_SE_PARM) {
				if(s->act->d.cons.cdr->type != STT_ST_CONS) return 5;
				if(prv->type == STT_ST_NIL) {
					s->act = s->out = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_symbol(s->stt, "BACKQUOTE"), stt_create_cons(s->stt, s->act->d.cons.cdr->d.cons.car, stt_create_nil(s->stt))), stt_create_nil(s->stt));
					return 0;
				} else if(prv->type == STT_ST_CONS) {
					prv->d.cons.cdr = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_symbol(s->stt, "BACKQUOTE"), stt_create_cons(s->stt, s->act->d.cons.cdr->d.cons.car, stt_create_nil(s->stt))), stt_create_nil(s->stt));
					s->act = prv;
					continue;
				} else return 6;
			}

			/* list special: , */
			if(code == STT_SE_WFUNC) {
				if(s->act->d.cons.cdr->type != STT_ST_CONS) return 7;
				if(prv->type == STT_ST_NIL) {
					s->act = s->out = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_symbol(s->stt, "UNQUOTE"), stt_create_cons(s->stt, s->act->d.cons.cdr->d.cons.car, stt_create_nil(s->stt))), stt_create_nil(s->stt));
					return 0;
				} else if(prv->type == STT_ST_CONS) {
					prv->d.cons.cdr = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_symbol(s->stt, "UNQUOTE"), stt_create_cons(s->stt, s->act->d.cons.cdr->d.cons.car, stt_create_nil(s->stt))), stt_create_nil(s->stt));
					s->act = prv;
					continue;
				} else return 8;
			}

			/* list special: ,@ */
			if(code == STT_SE_TYPE) {
				if(s->act->d.cons.cdr->type != STT_ST_CONS) return 9;
				if(prv->type == STT_ST_NIL) {
					s->act = s->out = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_symbol(s->stt, "UNQUOTE-SPLICING"), stt_create_cons(s->stt, s->act->d.cons.cdr->d.cons.car, stt_create_nil(s->stt))), stt_create_nil(s->stt));
					return 0;
				} else if(prv->type == STT_ST_CONS) {
					prv->d.cons.cdr = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_symbol(s->stt, "UNQUOTE-SPLICING"), stt_create_cons(s->stt, s->act->d.cons.cdr->d.cons.car, stt_create_nil(s->stt))), stt_create_nil(s->stt));
					s->act = prv;
					continue;
				} else return 10;
			}

			/* #N= label creation */
			if(code >= STT_SE_LAST) {
				if(s->act->d.cons.cdr->type != STT_ST_CONS) return 12;
				s->label = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_int(s->stt, code - STT_SE_LAST), s->act->d.cons.cdr->d.cons.car), s->label);

				if(prv->type == STT_ST_NIL) {
					s->act = s->out = stt_create_cons(s->stt, s->act->d.cons.cdr->d.cons.car, stt_create_nil(s->stt));
					return 0;
				} else if(prv->type == STT_ST_CONS) {
					prv->d.cons.cdr = stt_create_cons(s->stt, s->act->d.cons.cdr->d.cons.car, stt_create_nil(s->stt));
					s->act = prv;
					continue;
				} else return 13;
			}
		}

		s->act = s->act->d.cons.cdr;

		return 0;
	} else return 11;
}

/* test for symbol constituents */
static int issymbol(int ch) {
	return isalnum(ch) || (ispunct(ch) && ch != ':' && ch != ';' && ch != '#' && ch != '`' && ch != ',' && ch != '.'
			&& ch != '\'' && ch != '(' && ch != ')');
}

/* read a character */
stt_obj_t *stt_read_ch(stt_read_state_t *s, int ch)
{
	stt_obj_t *lst;
	char *c;
	int code;
	long num;

	if(s->ahead == 1) {
		/* this happens when the current character is to be reconsidered
		 * in a different state (thus the location should stay the same) */
		s->ahead = 0;
	} else {
		if(ch == '\n') {
			s->stt->parsed_line = ++ s->line;
			s->col = 1;
		} else s->col ++;
	}

	s->stt->lastcall_file = s->stt->file;
	s->stt->lastcall_line = s->stt->parsed_line;

	switch(s->state) {

		/* default state */
		case STT_RS_NORM:

			if(ch == '(') {
				if(s->act->type == STT_ST_CONS && s->act->d.cons.car->type == STT_ST_ERROR &&
					s->act->d.cons.car->d.error.code >= STT_SE_LAST) {

					/* have to assign the first cons of the new list to a label */
					s->stacklabel = s->act->d.cons.car->d.error.code - STT_SE_LAST;
					s->act = s->act->d.cons.car->d.error.obj;
					if(s->act->type == STT_ST_CONS)
						s->act->d.cons.cdr = stt_create_nil(s->stt);
				}

				s->stack = stt_create_cons(s->stt, stt_create_cons(s->stt, s->out, s->act), s->stack);
				s->out = s->act = stt_create_nil(s->stt);
			} else if(ch == ')') {
				if(s->stacklabel > -1) {
					if(s->act->type == STT_ST_NIL) {
						/* bind label to empty list */
						s->label = stt_create_cons(s->stt, stt_create_cons(s->stt, stt_create_int(s->stt, s->stacklabel), s->act), s->label);
						s->stacklabel = -1;
					}
				}

				if(s->stack->type != STT_ST_CONS)
					return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected ')'");
				lst = s->out;
				if(s->act->type == STT_ST_CONS) {
					if(s->act->d.cons.car->type == STT_ST_ERROR)
						return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Unexpected ')' following list special",
								stt_create_int(s->stt, s->act->d.cons.car->d.error.code));
				}

				s->out = s->stack->d.cons.car->d.cons.car;
				s->act = s->stack->d.cons.car->d.cons.cdr;
				s->stack = s->stack->d.cons.cdr;
				if((code = read_append(s, lst)))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}
			} else if(isspace(ch)) {
			} else if(ch == '#') {
				s->state = STT_RS_HASH;
				return NULL;
			} else if(ch == ';') {
				s->state = STT_RS_COMMENT;
				return NULL;
			} else if(ch == '\"') {
				s->state = STT_RS_STRING;
				return NULL;
			} else if(ch == EOF) {
				if(s->stack->type != STT_ST_NIL)
					return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");
				return NULL;
			} else if(isdigit(ch)) {
				s->state = STT_RS_INTEGER;
				token_appch(s, ch);
				return NULL;
			} else if(ch == '-') {
				s->state = STT_RS_MINUS;
				token_appch(s, ch);
				return NULL;
			} else if(issymbol(ch)) {
				s->state = STT_RS_SYMBOL;
				token_appch(s, ch);
				return NULL;
			} else if(ch == ':') {
				s->state = STT_RS_COLON;
				return NULL;
			} else if(ch == '.') {
				s->state = STT_RS_DOT;
				return NULL;
			} else if(ch == '\'') {
				if((code = read_append(s, stt_create_error_with_obj(s->stt, STT_SE_UNDEF, NULL, s->act))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				return NULL;
			} else if(ch == '`') {
				if((code = read_append(s, stt_create_error_with_obj(s->stt, STT_SE_PARM, NULL, s->act))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				return NULL;
			} else if(ch == ',') {
				s->state = STT_RS_COMMA;
				return NULL;
			} else {
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected character");
			}

			break;

		/* after a - */
		case STT_RS_MINUS:

			if(isdigit(ch)) {
				token_appch(s, ch);
				s->state = STT_RS_INTEGER;
				return NULL;
			} else if(issymbol(ch)) {
				token_appch(s, ch);
				s->state = STT_RS_SYMBOL;
				return NULL;
			} else if(ch == '.') {
				token_appch(s, ch);
				s->state = STT_RS_NUM;
				return NULL;
			} else {
				if((code = read_append(s, stt_create_symbol(s->stt, "-"))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				resettoken(s);
				s->state = STT_RS_NORM;

				s->ahead = 1;
				lst = stt_read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}

				return NULL;
			}

			break;

		/* reading an integer */
		case STT_RS_INTEGER:

			if(isdigit(ch)) {
				token_appch(s, ch);
			} else if(ch == '.') {
				token_appch(s, ch);
				s->state = STT_RS_NUM;
				return NULL;
			} else if(issymbol(ch)) {
				token_appch(s, ch);
				s->state = STT_RS_SYMBOL;
				return NULL;
			} else {
				if((code = read_append(s, stt_create_int(s->stt, strtol(s->token, NULL, 10)))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				resettoken(s);
				s->state = STT_RS_NORM;

				s->ahead = 1;
				lst = stt_read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}

				return NULL;
			}

			break;

		/* reading a float */
		case STT_RS_NUM:

			if(isdigit(ch)) {
				token_appch(s, ch);
			} else {
				if((code = read_append(s, stt_create_num(s->stt, atof(s->token)))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				resettoken(s);
				s->state = STT_RS_NORM;

				s->ahead = 1;
				lst = stt_read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}

				return NULL;
			}

			break;

		/* after a : */
		case STT_RS_COLON:
			if(issymbol(ch)) {
				token_appch(s, ch);
			} else {
				if(!s->token)
					return stt_create_error(s->stt, STT_SE_PARSE, "Invalid ':'");
				for(c = s->token; *c; c++)
					if(!isdigit(*c)) break;
				if(!*c)
					return stt_create_error(s->stt, STT_SE_PARSE, "Invalid ':'");
				lst = stt_create_new_symbol(s->stt, s->token);
				lst->d.symbol.colon = 1;
				if((code = read_append(s, lst)))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				resettoken(s);
				s->state = STT_RS_NORM;

				s->ahead = 1;
				lst = stt_read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}

				return NULL;
			}

			break;

		/* reading a symbol */
		case STT_RS_SYMBOL:

			if(issymbol(ch)) {
				token_appch(s, ch);
			} else {
				if((code = read_append(s, stt_create_symbol(s->stt, s->token))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				resettoken(s);
				s->state = STT_RS_NORM;

				s->ahead = 1;
				lst = stt_read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}

				return NULL;
			}

			break;

		/* after #\ */
		case STT_RS_CHAR_CONST:
			if((code = read_append(s, stt_create_char(s->stt, ch))))
				return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
			s->state = STT_RS_NORM;

			if(s->stack->type == STT_ST_NIL) {
				lst = s->out->d.cons.car;
				s->out = s->act = stt_create_nil(s->stt);
				s->label = stt_create_nil(s->stt);
				return lst;
			}

			return NULL;

		/* after a # */
		case STT_RS_HASH:
			if(isdigit(ch)) {
				token_appch(s, ch);
				s->state = STT_RS_HASH_INT;
				return NULL;
			}

			switch(ch) {
			case '\\':
				s->state = STT_RS_CHAR_CONST;
				return NULL;
			case '"':
				s->state = STT_RS_RAW_DBL;
				return NULL;
			case '\'':
				s->state = STT_RS_RAW_QUOT;
				return NULL;
			case '!':
				s->state = STT_RS_COMMENT;
				return NULL;
			default:
				return stt_create_error(s->stt, STT_SE_PARSE, "Invalid stuff after #");
			}

		/* after #[0-9]+ */
		case STT_RS_HASH_INT:
			if(isdigit(ch)) {
				token_appch(s, ch);
				return NULL;
			}


			switch(ch) {
			case '=':
				num = atoi(s->token);
				resettoken(s);
				/* insert label */
				if((code = read_append(s, stt_create_error_with_obj(s->stt, STT_SE_LAST + num, NULL, s->act))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				s->state = STT_RS_NORM;
				return NULL;
			case '#':
				num = strtol(s->token, NULL, 10);
				resettoken(s);
				/* invoke label */
				{
					stt_obj_t *c = s->label;
					for(; c && c->type == STT_ST_CONS; c = c->d.cons.cdr) {
						assert(c->d.cons.car->type == STT_ST_CONS);
						assert(c->d.cons.car->d.cons.car->type == STT_ST_INT);
						if(num == c->d.cons.car->d.cons.car->d.integer.value) {
							if((code = read_append(s, c->d.cons.car->d.cons.cdr)))
								return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
							s->state = STT_RS_NORM;

							if(s->stack->type == STT_ST_NIL) {
								lst = s->out->d.cons.car;
								s->out = s->act = stt_create_nil(s->stt);
								s->label = stt_create_nil(s->stt);
								return lst;
							}

							return NULL;
						}
					}
				}
				return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Invoked label not found", stt_create_int(s->stt, num));
			default:
				return stt_create_error(s->stt, STT_SE_PARSE, "Invalid stuff after #[0-9]+");
			}

		/* filename-reseting comment (;%#) */
		case STT_RS_COMMENT_FILE:
			if(ch != '\n' && isspace(ch) && !s->token) break;
			if(ch == ' ' || ch == ';' || ch == '\n') {
				if(s->token) {
					stt_mark_miscfree(s->stt, s->stt->file = stt_strdup(s->token));
					s->line = 0;
					resettoken(s);
				}
				s->state = (ch == '\n' ? STT_RS_NORM : STT_RS_COMMENT_BODY);
				return NULL;
			} else token_appch(s, ch);
			break;

		/* after a ;% */
		case STT_RS_COMMENT_1:
			if(ch == '#') {
				s->state = STT_RS_COMMENT_FILE;
				return NULL;
			}

			if(ch == '\n') {
				s->state = STT_RS_NORM;
				return NULL;
			}

			s->state = STT_RS_COMMENT_BODY;
			return NULL;

		/* inside a comment */
		case STT_RS_COMMENT_BODY:
			if(ch == '\n') {
				s->state = STT_RS_NORM;
				return NULL;
			}
			break;

		/* beginning a comment */
		case STT_RS_COMMENT:
			if(ch == ';') break;
			if(ch == '\n') {
				s->state = STT_RS_NORM;
				return NULL;
			}
			if(isspace(ch)) break;
			if(ch == '%') {
				s->state = STT_RS_COMMENT_1;
				return NULL;
			}
			s->state = STT_RS_COMMENT_BODY;
			return NULL;

		/* reading a string */
		case STT_RS_STRING:

			if(ch == '\\') {
				s->state = STT_RS_STRING_ESCAPE;
				return NULL;
			} else if(ch == '\"') {
				if((code = read_append(s, stt_create_string(s->stt, s->token ? s->token : ""))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				resettoken(s);
				s->state = STT_RS_NORM;

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}

				return NULL;
			} else if(ch == EOF) {
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");
			}

			token_appch(s, ch);
			break;

		/* after a \ while reading a string */
		case STT_RS_STRING_ESCAPE:

			if(ch == EOF)
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");

			switch(ch) {
				case 'a': ch = '\a'; break;
				case 'b': ch = '\b'; break;
				case 'f': ch = '\f'; break;
				case 'n': ch = '\n'; break;
				case 'r': ch = '\r'; break;
				case 't': ch = '\t'; break;
				case 'v': ch = '\v'; break;
				case 'x':
					/* supporting \xHH style character input */
					s->state = STT_RS_STRING_ESCAPE_HEX;
					return NULL;
				default: ;
			}

			token_appch(s, ch);
			s->state = STT_RS_STRING;
			return NULL;

		/* \xHH character input, first H */
		case STT_RS_STRING_ESCAPE_HEX:

			if(ch == EOF)
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");

			if(ch >= '0' && ch <= '9')
				s->ch = (ch - '0') << 4;
			else if(ch >= 'a' && ch <= 'f')
				s->ch = (ch - 'a' + 10) << 4;
			else if(ch >= 'A' && ch <= 'F')
				s->ch = (ch - 'A' + 10) << 4;
			else
				return stt_create_error(s->stt, STT_SE_PARSE, "Invalid first H in \\xHH escaped character");

			s->state = STT_RS_STRING_ESCAPE_HEX_2;
			return NULL;

		/* \xHH character input, second H */
		case STT_RS_STRING_ESCAPE_HEX_2:

			if(ch == EOF)
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");

			if(ch >= '0' && ch <= '9')
				s->ch += (ch - '0');
			else if(ch >= 'a' && ch <= 'f')
				s->ch += (ch - 'a' + 10);
			else if(ch >= 'A' && ch <= 'F')
				s->ch += (ch - 'A' + 10);
			else
				return stt_create_error(s->stt, STT_SE_PARSE, "Invalid second H in \\xHH escaped character");

			token_appch(s, s->ch);
			s->state = STT_RS_STRING;
			return NULL;

		/* after a . */
		case STT_RS_DOT:

			if(isdigit(ch)) {
				token_appch(s, '.');
				token_appch(s, ch);
				s->state = STT_RS_NUM;
				return NULL;
			} else {
				s->state = STT_RS_NORM;
				if((code = read_append(s, stt_create_error_with_obj(s->stt, STT_SE_WTF, NULL, s->act))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
	
				s->ahead = 1;
				return stt_read_ch(s, ch);
			}

			break;

		/* after a , */
		case STT_RS_COMMA:

			if(ch == '@') {
				s->state = STT_RS_NORM;
				if((code = read_append(s, stt_create_error_with_obj(s->stt, STT_SE_TYPE, NULL, s->act))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				return NULL;
			} else if(ch == EOF) {
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");
			} else {
				s->state = STT_RS_NORM;
				if((code = read_append(s, stt_create_error_with_obj(s->stt, STT_SE_WFUNC, NULL, s->act))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				s->ahead = 1;
				return stt_read_ch(s, ch);
			}

			break;

		/* in a #" raw string literal */
		case STT_RS_RAW_DBL:
		
			if(ch == '\\') {
				s->state = STT_RS_RAW_DBL_ESC;
				token_appch(s, ch);
				return NULL;
			} else if(ch == '\"') {
				if((code = read_append(s, stt_create_string(s->stt, s->token ? s->token : ""))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				resettoken(s);
				s->state = STT_RS_NORM;

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}

				return NULL;
			} else if(ch == EOF) {
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");
			}

			token_appch(s, ch);
			break;

		/* in a #' raw string literal */
		case STT_RS_RAW_QUOT:

			if(ch == '\\') {
				s->state = STT_RS_RAW_QUOT_ESC;
				token_appch(s, ch);
				return NULL;
			} else if(ch == '\'') {
				if((code = read_append(s, stt_create_string(s->stt, s->token ? s->token : ""))))
					return stt_create_error_with_obj(s->stt, STT_SE_PARSE, "Dot confusion", stt_create_int(s->stt, code));
				resettoken(s);
				s->state = STT_RS_NORM;

				if(s->stack->type == STT_ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = stt_create_nil(s->stt);
					s->label = stt_create_nil(s->stt);
					return lst;
				}

				return NULL;
			} else if(ch == EOF) {
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");
			}

			token_appch(s, ch);
			break;

		/* in a #" raw string literal, after a \ */
		case STT_RS_RAW_DBL_ESC:
			
			if(ch == EOF)
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");

			token_appch(s, ch);
			s->state = STT_RS_RAW_DBL;
			return NULL;

		/* in a #' raw string literal, after a \ */
		case STT_RS_RAW_QUOT_ESC:

			if(ch == EOF)
				return stt_create_error(s->stt, STT_SE_PARSE, "Unexpected EOF");

			token_appch(s, ch);
			s->state = STT_RS_RAW_QUOT;
			return NULL;

		default:
			return stt_create_error(s->stt, STT_SE_PARSE, "Undefined state");

	}

	return NULL;
}
