#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef READLINE
# include <readline/readline.h>
# include <readline/history.h>
#endif /* READLINE */
#include <estutter/stutter.h>
#include <estutter/compat.h>
#include <assert.h>

#ifdef WITH_LOCALE
#include <locale.h>
#endif

stt_ctx_t stt;
stt_varctx_t *global;

void main_uninit(void)
{
	stt_uninit(&stt);
}

/* FIXME: remove getopt */
extern int optind;
extern char *optarg;
extern int getopt(int argc, char *const argv[], const char *optstring);

/* No readline support on WinCE.
 * FIXME: no console support, either. --dig */
#ifdef READLINE

stt_varctx_t *completion_ctx;

char * word_generator(const char *text, int state)
{
	static int slot, len;
	static stt_var_t *elem;
	char *name;

	if(!state) {
		slot = 0;
		elem = NULL;
		len = strlen(text);
	}

	while(slot < completion_ctx->ns) {
		if(!elem) elem = completion_ctx->hash[slot];
		if(!elem) {
			slot++;
			continue;
		}

		name = stt_get_symbolname(&stt, elem->sdbm, elem->id);
		elem = elem->next;
		if(!elem) slot++;
		if(strncasecmp(name, text, len) == 0)
			return stt_strdup(name);
	}

	return NULL;
}

char ** stutter_completion(const char *text, int start, int end)
{
	return rl_completion_matches(text, word_generator);
}

void shell()
{
	stt_obj_t *obj;
	stt_s_mark mark;
	char *buf, *c;
	stt_read_state_t *rs;

	char *hist;
	int histbuf, histlen;
	int beg = 1;

	completion_ctx = global;

	/* set up readline */
	rl_readline_name = "stutter";
	rl_attempted_completion_function = stutter_completion;

	rs = stt_create_read_state(&stt);

	histbuf = 256;
	hist = (char *) malloc(histbuf);
	assert(hist != NULL);
	histlen = 0;
	hist[0] = 0;

	printf("Welcome to Embedded STUTTER %s\n\n", stt_version);

	stt_varctx_set(global, "*ARGV*", stt_create_nil(&stt));
	stt_varctx_set(global, "$0", stt_create_string(&stt, "shell"));

	while((buf = readline(rs->stack->type != STT_ST_NIL ? "" : "> "))) {
		obj = NULL;
		for(c = buf; ; c++) {
			if(*c && beg && !isspace(*c))
				beg = 0;

			if(!beg) {
				if(histlen == histbuf - 1) {
					hist = (char *) realloc(hist, histbuf *= 2);
					assert(hist != NULL);
				}
				hist[histlen] = *c ? *c : '\n';
				hist[++histlen] = 0;
			}

			if(*c) {
				obj = stt_read_ch(rs, *c);
			} else {
				obj = stt_read_ch(rs, '\n');
			}
			if(obj) {
				if(obj->type == STT_ST_ERROR && obj->d.error.code == STT_SE_PARSE) {
					for(c++; *c; c++) {
						if(histlen == histbuf - 1) {
							hist = (char *) realloc(hist, histbuf *= 2);
							assert(hist != NULL);
						}
						hist[histlen] = *c ? *c : '\n';
						hist[++histlen] = 0;
					}

					add_history(hist);
					histlen = 0;
					stt_printobj(global, obj);
					stt_stream_stdout_write(global, '\n');
					stt_read_reset(rs);
					break;
				}

				add_history(hist);
				histlen = 0;
				if(obj->type == STT_ST_ERROR) break;
				mark = stt_gc_prot_mark(&stt);
				obj = stt_eval(global, obj);
				stt_printobj(global, obj);
				stt_stream_stdout_write(global, '\n');
				stt_gc_prot_free(&stt, mark, NULL);
				obj = NULL;
				beg = 1;
			}
			if(!*c) break;
		}
	}

	free(hist);
	stt_free_read_state(rs);
	stt_varctx_free(global);
}

#else  /* READLINE */

/* this shell() reads from stdin */
void shell()
{
	stt_obj_t *obj;
	stt_s_mark mark;
	int c;
	stt_read_state_t *rs;

	rs = stt_create_read_state(&stt);

	printf("Welcome to Embedded STUTTER %s\n\n> ", stt_version);

	stt_varctx_set(global, "*ARGV*", stt_create_nil());
	stt_varctx_set(global, "*NO-READLINE*", stt_create_t());
	stt_varctx_set(global, "$0", stt_create_string("shell"));

	while((c = getchar())) {
		obj = stt_read_ch(rs, c);

		if(obj) {
			if(obj->type == STT_ST_ERROR) {
				stt_fp_printobj(stderr, global, obj);
				fputc('\n', stderr);
				stt_read_reset(rs);
			} else {
				mark = stt_gc_prot_mark();
				obj = stt_eval(global, obj);
				stt_printobj(global, obj);
				putchar('\n');
				stt_gc_prot_free(mark, NULL);
			}
		}

		if(c == '\n') fputs("> ", stdout);
		if(c == EOF) break;
	}

	stt_free_read_state(rs);
	stt_varctx_free(global);
}


#endif /* READLINE */

int main(int argc, char *argv[])
{
	int res, i, sh = 1;
	stt_obj_t *obj;
	stt_obj_t *sargv;

#ifdef WITH_LOCALE
	setlocale(LC_CTYPE, "");
#endif

	stt_init(&stt);
	atexit(main_uninit);

	global = stt_varctx_create(&stt, NULL, 1024);
	stt_register_builtins(global);
	stt_register_stdio(global);
	obj = stt_load_system(global);
	if(obj->type == STT_ST_ERROR) {
		stt_fp_printobj(stderr, global, obj);
		fputc('\n', stderr);
	}

	while((res = getopt(argc, argv, "hvG:gel:")) != -1) {
		if(res == '?' || res == 'h') {
			printf("estt %s\n\n", stt_version);
			printf(" -v       print interpreter version\n");
			printf(" -h       print this help screen\n");
			printf(" -G <n>   set auto-GC interval to per N objects allocated\n");
			printf(" -g       verbose GC\n");
			printf(" -e       disable readline interface (parse from stdin)\n");
			printf(" -l <f>   read a file before startup\n");
			return 0;
		}

		if(res == 'e')
			sh = 0;
		if(res == 'g')
			stt.gc_verbose = 1;
		if(res == 'G') {
			stt.gc_alloc_limit = atoi(optarg);
			if(stt.gc_alloc_limit < 0) {
				stt.gc_alloc_limit = 0;
				printf("GC_ALLOC_LIMIT cannot be smaller than 0\n");
			}
		}

		if(res == 'v') {
			printf("%s\n", stt_version);
			return 0;
		}

		if(res == 'l') {
			stt_obj_t *obj = stt_create_string(&stt, optarg);
			obj = STT_LIST2(&stt, STT_SYM(&stt, "LOAD"), obj);
			obj = stt_eval(global, obj);
			if(obj->type == STT_ST_ERROR) {
				stt_printobj(global, obj);
				stt_stream_stdout_write(global, '\n');
				stt_varctx_free(global);
				stt_gc_prot_free(&stt, 0, NULL);
				stt_gc_sweep(&stt, NULL);
				return 1;
			}
		}
	}

	if(optind == argc && sh) {
		shell();
		return 0;
	}

	if(optind == argc) {
		stt_varctx_set(global, "*ARGV*", stt_create_nil(&stt));
		stt_varctx_set(global, "$0", stt_create_string(&stt, "stdin"));
		obj = stt_evalfp(global, stdin);
		if(obj->type == STT_ST_ERROR) {
			stt_fp_printobj(stderr, global, obj);
			fputc('\n', stderr);
		}
	} else {
		sargv = stt_create_nil(&stt);
		obj = NULL;
		for(i=optind+1; i<argc; i++) {
			if(!obj) {
				obj = sargv = stt_create_cons(&stt, stt_create_string(&stt, argv[i]), stt_create_nil(&stt));
			} else {
				obj->d.cons.cdr = stt_create_cons(&stt, stt_create_string(&stt, argv[i]), stt_create_nil(&stt));
				obj = obj->d.cons.cdr;
			}
		}
		stt_varctx_set(global, "*ARGV*", sargv);

		if(optind < argc) {
			stt_obj_t *obj = stt_create_string(&stt, argv[optind]);
			stt_varctx_set(global, "$0", obj);
			obj = STT_LIST2(&stt, stt_create_symbol(&stt, "LOAD"), obj);
			obj = stt_eval(global, obj);
			if(obj->type == STT_ST_ERROR) {
				stt_printobj(global, obj);
				stt_stream_stdout_write(global, '\n');
				stt_varctx_free(global);
				stt_gc_prot_free(&stt, 0, NULL);
				stt_gc_sweep(&stt, NULL);
				return 1;
			}
		}
	}

	stt_varctx_free(global);

	return 0;
}

